/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AuditService.java 19729 2010-04-22 13:30:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.admin.auditconsole.service;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.regex.PatternSyntaxException;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.ow2.jonas.lib.management.extensions.manager.ManagementEntryPoint;
import org.ow2.util.auditreport.api.AuditorJMXObjectNames;
import org.ow2.util.auditreport.impl.util.ObjectEncoder;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Listener for audit notifications.
 */
public class AuditService implements Serializable, NotificationListener {
    /**
     * Logger of the class
     */
    private Log logger = LogFactory.getLog(AuditService.class);

    /**
     * Serial Version UID.
     */
    private static final long serialVersionUID = 1L;

    /**
     * Reports waiting to be displayed.
     */
    private ArrayList reportsObjects = new ArrayList();

    /**
     * The EZB JMX server.
     */
    private transient MBeanServerConnection serv = null;

    /**
     * Mbean name of EZB auditor.
     */
    private ObjectName ejbNotifier = null;

    /**
     * Mbean name of web auditor.
     */
    private ObjectName webNotifier = null;

    /**
     * Mbean name of jndi auditor.
     */
    private ObjectName jndiNotifier = null;

    /**
     * MBean name of the logger.
     */
    private ObjectName loggerMBean = null;

    /**
     * Number of invocations detected.
     */
    private int invocations = 0;

    /**
     * Domain name.
     */
    private String domainName = null;

    /**
     * Constructor.
     */
    public AuditService() {
        init();
    }

    /**
     * Init the notifications listening.
     */
    private void init() {
        reportsObjects = new ArrayList();
        try {
            domainName = ManagementEntryPoint.getInstance().getDomainName();
            serv = ManagementEntryPoint.getInstance().getServerConnection();
            ejbNotifier = new ObjectName(domainName + AuditorJMXObjectNames.EJBAUDITOR_TYPE_COMPONENT + ",name=EasyBeans");
            webNotifier = new ObjectName(domainName + AuditorJMXObjectNames.WEBAUDITOR_TYPE_COMPONENT + ",name=JOnAS");
            jndiNotifier = new ObjectName(domainName + AuditorJMXObjectNames.JNDIAUDITOR_TYPE_COMPONENT + ",name=JOnAS");

            loggerMBean = new ObjectName(domainName + AuditorJMXObjectNames.LOGGERAUDITOR_TYPE_COMPONENT + ",name=JOnAS");

        } catch (Exception ex) {
            logger.error("Error when creating notifs listeners : ", ex);
        }
    }

    /**
     * Start the listening.
     */
    public void startListening() {
        if (serv == null) {
            init();
        }
        getNewRegistrations();
        this.reportsObjects.clear();
    }

    /**
     * @return waiting reports.
     */
    public ArrayList getReportObjects() {

        ArrayList tmp = (ArrayList) this.reportsObjects.clone();
        this.reportsObjects.clear();
        return tmp;
    }

    /**
     * Start the audit on every auditor component.
     */
    public void startAuditing() {
        // TODO : start jonas auditing services .
    }

    /**
     * Stop the audit on every auditor component.
     */
    public void stopAuditing() {
        // TODO : stop jonas auditing services.
    }

    /**
     * This methode handle a JMX notification.
     *
     * @param notification the notification.
     * @param handback the handback.
     */
    public void handleNotification(final Notification notification, final Object handback) {
        this.reportsObjects.add(ObjectEncoder.decode(notification.getMessage()));
        invocations++;
    }

    /**
     * This method add a listener on a MBean if this MBean is present in the JMX
     * server.
     */
    public void getNewRegistrations() {
        logger.info("Start listening audit notifications");
        try {
            if (serv != null) {
                if (serv.isRegistered(ejbNotifier)) {
                    serv.addNotificationListener(ejbNotifier, this, null, null);
                }
                if (serv.isRegistered(webNotifier)) {
                    serv.addNotificationListener(webNotifier, this, null, null);
                }
                if (serv.isRegistered(jndiNotifier)) {
                    serv.addNotificationListener(jndiNotifier, this, null, null);
                }
            }
        } catch (Exception ex) {
            logger.error("Error at listeners registration : ", ex);
        }
    }

    /**
     * Unregister the JMX listener.
     */
    public void unregister() {
        logger.info("Stop listening audit notifications");
        try {
            if (serv.isRegistered(ejbNotifier)) {
                serv.removeNotificationListener(ejbNotifier, this);
            }
            if (serv.isRegistered(webNotifier)) {
                serv.removeNotificationListener(webNotifier, this);
            }
            if (serv.isRegistered(jndiNotifier)) {
                serv.removeNotificationListener(jndiNotifier, this);
            }
        } catch (Exception ex) {
            logger.error("Error at listeners unregistration : ", ex);
        }
    }

    /**
     * Get a Map with data name and data value.
     * @param id the NODEID on which we want to retrieve datas
     * @return the expected map
     * @throws IOException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     */
    @SuppressWarnings("unchecked")
    protected Map<String, Object> getDescriptionNodeId(final String id) throws IOException, ReflectionException,
        InstanceNotFoundException, MBeanException {
        Map<String, Object> map = null;
        if (serv != null) {
            if (serv.isRegistered(loggerMBean)) {
                map = (Map<String, Object>) serv.invoke(loggerMBean, "getDescriptionNodeId", new Object[] {id},
                    new String[] {String.class.getName()});
            }
        }
        // if N/A
        return map;
    }

    /**
     * Gets Node Info for the given node ID.
     * @param nodeID the expected Node ID
     * @return data for the given node
     * @throws IOException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     */
    protected NodeInfo getNodeInfo(final String nodeID) throws IOException, ReflectionException, InstanceNotFoundException,
        MBeanException {
        Map<String, Object> map = getDescriptionNodeId(nodeID);
        NodeInfo nodeInfo = null;
        if (map != null) {
            nodeInfo = new NodeInfo();
            nodeInfo.id = nodeID;
            nodeInfo.date = (Date) map.get("date");
            nodeInfo.name = (String) map.get("name");
            nodeInfo.application = (String) map.get("application");
            nodeInfo.desc = (String) map.get("desc");
            nodeInfo.type = (String) map.get("type");
            nodeInfo.depth = (Integer) map.get("depth");
            nodeInfo.localChildren = (String[]) map.get("children");
            nodeInfo.execTime = (Double) map.get("execTime");
        }
        return nodeInfo;

    }

    /**
     * Gets all available nodes at the root level (entry points)
     * @param keyword a keyword used for the search
     * @param maxResults the maximum results to send
     * @param depth the wanted depth
     * @return the node Info list if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws IOException if there is a failure
     */
    public List<NodeInfo> getTreeAvailableNodes(final String keyword, final int maxResults, final int depth)
        throws InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        List<NodeInfo> nodesInfo = new ArrayList<NodeInfo>();
        String[] nodesID = null;
        // Get list of nodes ID
        if (serv != null) {
            if (serv.isRegistered(loggerMBean)) {
                nodesID = (String[]) serv.invoke(loggerMBean, "getRootNodesID", null, null);
            }
        }
        // Nodes available
        if (nodesID != null) {
            // For each node, get a description
            for (String nodeID : nodesID) {
                // Break when we have the results
                if (nodesInfo.size() >= maxResults) {
                    break;
                }
                NodeInfo nodeInfo = getNodeInfo(nodeID);

                if (nodeInfo != null && nodeInfo.name != null) {
                    // Matching name ?
                    boolean matching = false;
                    try {
                        matching = nodeInfo.name.matches(keyword);
                    } catch (PatternSyntaxException e) {
                        matching = false;
                    }
                    if (!matching) {
                        continue;
                    }

                    if (nodeInfo.depth >= depth) {
                        nodesInfo.add(nodeInfo);
                    }

                }
            }
        }

        return nodesInfo;

    }

    /**
     * Build a new Element for the NODE Graph object.
     * @param document the document used to build the element
     * @param id the id of the node
     * @param nodeInfo the data for this node
     * @return a built node element
     */
    protected Element nodeElement(final Document document, final String id, final NodeInfo nodeInfo) {
        Element nodeElement = document.createElement("Node");
        nodeElement.setAttribute("id", id);
        nodeElement.setAttribute("name", nodeInfo.name);
        nodeElement.setAttribute("desc", nodeInfo.desc);
        nodeElement.setAttribute("nodeColor", "0xFF0000");

        String icon = "leaf";
        if ("EJB".equals(nodeInfo.type)) {
            icon = "10";
        } else if ("EJB_LOCAL".equals(nodeInfo.type)) {
            icon = "11";
        } else if ("EJB_REMOTE".equals(nodeInfo.type)) {
            icon = "12";
        } else if ("HTTP_GET".equals(nodeInfo.type)) {
            icon = "2";
        } else if ("HTTP_POST".equals(nodeInfo.type)) {
            icon = "2";
        }
        nodeElement.setAttribute("nodeIcon", icon);
        return nodeElement;
    }

    /**
     * Build a new Element for the EDGE Graph object.
     * @param document the document used to build the element
     * @param parentId the Parent id of the node (used to make a link between parent and current ID)
     * @param id the id of the node
     * @param label the label of this edge
     * @param flow the flow of this edge
     * @return a built edge element
     */
    protected Element edgeElement(final Document document, final String parentId, final String id, final String label, final String flow) {
        Element nodeElement = document.createElement("Edge");
        nodeElement.setAttribute("fromID", parentId);
        nodeElement.setAttribute("toID", id);
        nodeElement.setAttribute("edgeLabel", label);
        nodeElement.setAttribute("flow",flow);
        nodeElement.setAttribute("color","0x000000");

        return nodeElement;
    }

    /**
     * Complete the given document by adding XML nodes and edges.
     * @param document the XML document
     * @param root the entry node
     * @param first in order to know if we need to link with a parent or not
     * @param parentID the parent ID
     * @param id the current ID
     * @param nodeInfo the info about the node
     * @throws InstanceNotFoundException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws IOException if there is a failure
     */
    private void buildXML(final Document document, final Element root, final boolean first, final String parentID, final String id, final NodeInfo nodeInfo) throws InstanceNotFoundException, ReflectionException, MBeanException, IOException {

        // Link between parent and ourself
        if (!first) {
            root.appendChild(edgeElement(document, parentID, id, String.valueOf(nodeInfo.execTime) + " ms", "0"));
        }

        // Add the node
        root.appendChild(nodeElement(document, id, nodeInfo));

        String[] children = nodeInfo.localChildren;

        // For each children, add the XML data
        if (children != null) {
            int length = children.length;
            int i = 0;
            for (int c = length -1; c >= 0; c--) {
                String idCurrent = id + "_" + i;
                NodeInfo localNodeInfo = getNodeInfo(children[c]);
                buildXML(document, root, false, id, idCurrent, localNodeInfo);
                i++;
            }
        }

    }


    /**
     * Build an XML document and send it to the client
     * @param nodeID the root node ID
     * @return a document with Nodes and Edges
     * @throws InstanceNotFoundException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws IOException if there is a failure
     */
    public Document getGraphNodeID(final String nodeID) throws InstanceNotFoundException, ReflectionException, MBeanException, IOException {
        // Create builder with factory
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new IllegalStateException("Cannot build document builder", e);
        }

        // Create Document
        Document document = builder.newDocument();

        // Append root element
        Element root = document.createElement("Graph");
        document.appendChild(root);

        // Add root node


        buildXML(document, root, true, "", nodeID, getNodeInfo(nodeID));

        return document;

    }

    /**
     * @return number of invocations.
     */
    public int getInvocations() {
        return invocations;
    }
}
